@@ -13,9 +13,10 @@
 """Abstractions to interact with service models."""
 from collections import defaultdict
 
-from botocore.utils import CachedProperty, instance_cache
+from botocore.utils import CachedProperty, instance_cache, hyphenize_service_id
 from botocore.compat import OrderedDict
-
+from botocore.exceptions import MissingServiceIdError
+from botocore.exceptions import UndefinedModelAttributeError
 
 NOT_SET = object()
 
@@ -36,8 +37,9 @@ class InvalidShapeReferenceError(Exception):
     pass
 
 
-class UndefinedModelAttributeError(Exception):
-    pass
+class ServiceId(str):
+    def hyphenize(self):
+        return hyphenize_service_id(self)
 
 
 class Shape(object):
@@ -49,9 +51,11 @@ class Shape(object):
     SERIALIZED_ATTRS = ['locationName', 'queryName', 'flattened', 'location',
                         'payload', 'streaming', 'timestampFormat',
                         'xmlNamespace', 'resultWrapper', 'xmlAttribute',
-                        'jsonvalue']
+                        'eventstream', 'event', 'eventheader', 'eventpayload',
+                        'jsonvalue', 'timestampFormat', 'hostLabel']
     METADATA_ATTRS = ['required', 'min', 'max', 'sensitive', 'enum',
-                      'idempotencyToken', 'error', 'exception']
+                      'idempotencyToken', 'error', 'exception',
+                      'endpointdiscoveryid']
     MAP_TYPE = OrderedDict
 
     def __init__(self, shape_name, shape_model, shape_resolver=None):
@@ -105,6 +109,7 @@ class Shape(object):
             * resultWrapper
             * xmlAttribute
             * jsonvalue
+            * timestampFormat
 
         :rtype: dict
         :return: Serialization information about the shape.
@@ -162,6 +167,10 @@ class Shape(object):
         return "<%s(%s)>" % (self.__class__.__name__,
                              self.name)
 
+    @property
+    def event_stream_name(self):
+        return None
+
 
 class StructureShape(Shape):
     @CachedProperty
@@ -178,6 +187,13 @@ class StructureShape(Shape):
             shape_members[name] = self._resolve_shape_ref(shape_ref)
         return shape_members
 
+    @CachedProperty
+    def event_stream_name(self):
+        for member_name, member in self.members.items():
+            if member.serialization.get('eventstream'):
+                return member_name
+        return None
+
 
 class ListShape(Shape):
     @CachedProperty
@@ -282,6 +298,15 @@ class ServiceModel(object):
         else:
             return self.endpoint_prefix
 
+    @CachedProperty
+    def service_id(self):
+        try:
+            return ServiceId(self._get_metadata_property('serviceId'))
+        except UndefinedModelAttributeError:
+            raise MissingServiceIdError(
+                service_name=self._service_name
+            )
+
     @CachedProperty
     def signing_name(self):
         """The name to use when computing signatures.
@@ -306,12 +331,19 @@ class ServiceModel(object):
     def endpoint_prefix(self):
         return self._get_metadata_property('endpointPrefix')
 
+    @CachedProperty
+    def endpoint_discovery_operation(self):
+        for operation in self.operation_names:
+            model = self.operation_model(operation)
+            if model.is_endpoint_discovery_operation:
+                return model
+
     def _get_metadata_property(self, name):
         try:
             return self.metadata[name]
         except KeyError:
             raise UndefinedModelAttributeError(
-                '"%s" not defined in the metadata of the the model: %s' %
+                '"%s" not defined in the metadata of the model: %s' %
                 (name, self))
 
     # Signature version is one of the rare properties
@@ -328,6 +360,10 @@ class ServiceModel(object):
     def signature_version(self, value):
         self._signature_version = value
 
+    def __repr__(self):
+        return '%s(%s)' % (self.__class__.__name__, self.service_name)
+
+
 
 class OperationModel(object):
     def __init__(self, operation_model, service_model, name=None):
@@ -400,6 +436,16 @@ class OperationModel(object):
     def deprecated(self):
         return self._operation_model.get('deprecated', False)
 
+    @CachedProperty
+    def endpoint_discovery(self):
+        # Explicit None default. An empty dictionary for this trait means it is
+        # enabled but not required to be used.
+        return self._operation_model.get('endpointdiscovery', None)
+
+    @CachedProperty
+    def is_endpoint_discovery_operation(self):
+        return self._operation_model.get('endpointoperation', False)
+
     @CachedProperty
     def input_shape(self):
         if 'input' not in self._operation_model:
@@ -438,6 +484,33 @@ class OperationModel(object):
         shapes = self._operation_model.get("errors", [])
         return list(self._service_model.resolve_shape_ref(s) for s in shapes)
 
+    @CachedProperty
+    def endpoint(self):
+        return self._operation_model.get('endpoint')
+
+    @CachedProperty
+    def has_event_stream_input(self):
+        return self.get_event_stream_input() is not None
+
+    @CachedProperty
+    def has_event_stream_output(self):
+        return self.get_event_stream_output() is not None
+
+    def get_event_stream_input(self):
+        return self._get_event_stream(self.input_shape)
+
+    def get_event_stream_output(self):
+        return self._get_event_stream(self.output_shape)
+
+    def _get_event_stream(self, shape):
+        """Returns the event stream member's shape if any or None otherwise."""
+        if shape is None:
+            return None
+        event_name = shape.event_stream_name
+        if event_name:
+            return shape.members[event_name]
+        return None
+
     @CachedProperty
     def has_streaming_input(self):
         return self.get_streaming_input() is not None
@@ -650,8 +723,9 @@ class DenormalizedStructureBuilder(object):
         }
         if 'documentation' in model:
             shape['documentation'] = model['documentation']
-        if 'enum' in model:
-            shape['enum'] = model['enum']
+        for attr in Shape.METADATA_ATTRS:
+            if attr in model:
+                shape[attr] = model[attr]
         return shape
 
     def _build_scalar(self, model):
